import websocket
import json
import hashlib
import random
import threading
import hmac
import logging
import base64

from time import sleep
from types import SimpleNamespace

_LOGGER = logging.getLogger(__name__)

class LoadMode():
    """Wrapper Class to represent the Load Mode of the Wattpilot"""
    DEFAULT=3
    ECO=4
    NEXTTRIP=5


class Wattpilot(object):

    carValues = {}
    alwValues = {}
    astValues = {}
    lmoValues = {}
    ustValues = {}
    errValues = {}
    acsValues = {}

    lmoValues[3] = "Default"
    lmoValues[4] = "Eco"
    lmoValues[5] = "Next Trip"

    astValues[0] = "open"
    astValues[1] = "locked"
    astValues[2] = "auto"

    carValues[1] = "no car"
    carValues[2] = "charging"
    carValues[3] = "ready"
    carValues[4] = "complete"

    alwValues[0] = False
    alwValues[1] = True

    ustValues[0] = "Normal"
    ustValues[1] = "AutoUnlock"
    ustValues[2] = "AlwaysLock"

    errValues[0] = "Unknown Error"
    errValues[1] = "Idle"
    errValues[2] = "Charging"
    errValues[3] = "Wait Car"
    errValues[4] = "Complete"
    errValues[5] = "Error"

    acsValues[0] = "Open"
    acsValues[1] = "Wait"


    @property
    def allProps(self):
        """Returns a dictionary with all properties"""
        return self._allProps

    @property
    def allPropsInitialized(self):
        """Returns true, if all properties have been initialized"""
        return self._allPropsInitialized

    @property
    def cableType(self):
        """Returns the Cable Type (Ampere) of the connected cable"""
        return self._cableType

    @property
    def frequency(self):
        """Returns the power frequency"""
        return self._frequency

    @property
    def phases(self):
        """returns the phases"""
        return self._phases
    
    @property
    def energyCounterSinceStart(self):
        """Returns used kwh since start of charging"""
        return self._energyCounterSinceStart
    
    @property
    def errorState(self):
        """Returns error State"""
        return self._errorState

    @property
    def cableLock(self):
        return self._cableLock
    
    @property
    def energyCounterTotal(self):
        return self._energyCounterTotal

    @property
    def serial(self):
        """Returns the serial number of Wattpilot Device (read only)"""
        return self._serial
    @serial.setter
    def serial(self,value):
        self._serial = value
        if (self._password is not None) & (self._serial is not None):
            self._hashedpassword = base64.b64encode(hashlib.pbkdf2_hmac('sha512',self._password.encode(),self._serial.encode(),100000,256))[:32]
           
    @property
    def name(self):
        """Returns the name of Wattpilot Device (read only)"""
        return self._name


    @property
    def hostname(self):
        """Returns the DNS Hostname of Wattpilot Device (read only)"""
        return self._hostname

    @property
    def friendlyName(self):
        """Returns the friendly name of Wattpilot Device (read only)"""
        return self._friendlyName

    @property
    def manufacturer(self):
        """Returns the Manufacturer of Wattpilot Device (read only)"""
        return self._manufacturer

    @property
    def devicetype(self):
        return self._devicetype

    @property
    def protocol(self):
        return self._protocol
    
    @property
    def secured(self):
        return self._secured

    @property
    def password(self):
        return self._password
    @password.setter
    def password(self,value):
        self._password = value
        if (self._password is not None) & (self._serial is not None):
            self._hashedpassword = base64.b64encode(hashlib.pbkdf2_hmac('sha512',self._password.encode(),self._serial.encode(),100000,256))[:32]


    @property
    def url(self):
        return self._url
    @url.setter
    def url(self,value):
        self._url = value

    @property
    def connected(self):
        return self._connected

    @property
    def voltage1(self):
        return self._voltage1

    @property
    def voltage2(self):
        return self._voltage2

    @property
    def voltage3(self):
        return self._voltage3

    @property
    def voltageN(self):
        return self._voltageN

    @property
    def amps1(self):
        return self._amps1

    @property
    def amps2(self):
        return self._amps2

    @property
    def amps3(self):
        return self._amps3

    @property
    def power1(self):
        return self._power1

    @property
    def power2(self):
        return self._power2

    @property
    def power3(self):
        return self._power3

    @property
    def powerN(self):
        return self._powerN

    @property
    def power(self):
        return self._power

    @property
    def version(self):
        return self._version

    @property
    def amp(self):
        return self._amp

    @property
    def AccessState(self):
        return self._AccessState

    @property
    def firmware(self):
        """Returns the Firmwareversion of Wattpilot Device (read only)"""
        return self._firmware

    @property
    def WifiSSID(self):
        """Returns the SSID of the Wifi network currently connected (read only)"""
        return self._WifiSSID

    @property
    def AllowCharging(self):
        return self._AllowCharging

    @property
    def mode(self):
        return self._mode

    @property
    def carConnected(self):
        return self._carConnected

    @property
    def cae(self):
        """Returns true if Cloud API Access is enabled (read only)"""
        return self._cae

    @property
    def cak(self):
        """Returns the API Key for Cloud API Access (read only)"""
        return self._cak


    def __str__(self):
        """Returns a String representation of the core Wattpilot attributes"""
        if self.connected:
            ret = "Wattpilot: " + str(self.name) + "\n"
            ret = ret +  "Serial: " + str(self.serial) + "\n"
            ret = ret +  "Connected: " + str(self.connected) + "\n"
            ret = ret + "Car Connected: " + str(self.carConnected) + "\n"
            ret = ret + "Charge Status " + str(self.AllowCharging) + "\n"
            ret = ret + "Mode: " + str(self.mode) + "\n"
            ret = ret + "Power: " + str(self.amp) + "\n"
            ret = ret +  "Charge: " + "%.2f" % self.power + "kW" + " ---- " + str(self.voltage1) + "V/" + str(self.voltage2) + "V/" + str(self.voltage3) + "V" + " -- "
            ret = ret + "%.2f" % self.amps1 + "A/" + "%.2f" % self.amps2 + "A/" + "%.2f" % self.amps3 + "A" + " -- "
            ret = ret + "%.2f" % self.power1 + "kW/" + "%.2f" % self.power2 + "kW/" + "%.2f" % self.power3 + "kW" + "\n"
        else:
            ret = "Not connected"

        return ret
    def connect(self):
        self._wst = threading.Thread(target=self._wsapp.run_forever)
        self._wst.daemon = True
        self._wst.start()
        
        _LOGGER.info("Wattpilot connected")

    def register_message_callback(self,callback_fn):
        """signature of callback_fn: (wsapp,msg)"""
        self._message_callback = callback_fn

    def unregister_message_callback(self):
        self._message_callback = None

    def register_property_callback(self,callback_fn):
        """signature of callback_fn: (name,value)"""
        self._property_callback = callback_fn

    def unregister_property_callback(self):
        self._property_callback = None

    def set_power(self,power):
        self.send_update("amp",power)

    def set_mode(self,mode):
        self.send_update("lmo",mode)


    def send_update(self,name,value):
        message = {}
        message["type"]="setValue"
        self.__requestid = self.__requestid+1
        message["requestId"]=self.__requestid
        message["key"]=name
        message["value"]=value
        if (self._secured is not None):
            if  (self._secured > 0):
                self.__send(message,True)
            else:
                self.__send(message)
        else:
            self.__send(message)

    def __update_property(self,name,value):

        self._allProps[name] = value
        if name=="acs":
            self._AccessState = Wattpilot.acsValues[value]

        if name=="cbl":
            self._cableType = value

        if name=="fhz":
            self._frequency = value

        if name=="pha":
            self._phases = value
        
        if name=="wh":
            self._energyCounterSinceStart = value

        if name=="err":
            self._errorState = Wattpilot.errValues[value]

        if name=="ust":
            self._cableLock = Wattpilot.ustValues[value]

        if name=="eto":
            self._energyCounterTotal = value

        if name=="cae":
            self._cae = value
        if name=="cak":
            self._cak = value
        if name=="lmo":
            self._mode = Wattpilot.lmoValues[value]
        if name=="car":
            self._carConnected = Wattpilot.carValues[value]
        if name=="alw":
            self._AllowCharging = Wattpilot.alwValues[value]
        if name=="nrg":
            self._voltage1=value[0]
            self._voltage2=value[1]
            self._voltage3=value[2]
            self._voltageN=value[3]
            self._amps1=value[4]
            self._amps2=value[5]
            self._amps3=value[6]
            self._power1=value[7]*0.001
            self._power2=value[8]*0.001
            self._power3=value[9]*0.001
            self._powerN=value[10]*0.001
            self._power=value[11]*0.001
        if name=="amp":
            self._amp = value
        if name=="version":
            self._version = value
        if name=="ast":
            self._AllowCharging = self._astValues[value]
        if name=="fwv":
            self._firmware = value
        if name=="wss":
            self._WifiSSID=value
        if name=="upd":
            if value=="0":
                self._updateAvailable = False
            else:
                self._updateAvailable = True
        if self._property_callback != None:
            self._property_callback(name,value)

    def __on_hello(self,message):
        _LOGGER.info("Connected to WattPilot Serial %s",message.serial)
        if hasattr(message,"hostname"):
            self._name=message.hostname
        self.serial = message.serial
        if hasattr(message,"hostname"):
            self._hostname=message.hostname
        if hasattr(message,"version"):
            self._version=message.version
        self._manufacturer=message.manufacturer
        self._devicetype=message.devicetype
        self._protocol=message.protocol
        if hasattr(message,"secured"):
            self._secured=message.secured

    def __on_auth(self,wsapp,message):
        ran = random.randrange(10**80)
        self._token3 = "%064x" % ran
        self._token3 = self._token3[:32]
        hash1 = hashlib.sha256((message.token1.encode()+self._hashedpassword)).hexdigest()
        hash = hashlib.sha256((self._token3 + message.token2+hash1).encode()).hexdigest()
        response = {}
        response["type"] = "auth"
        response["token3"] = self._token3
        response["hash"] = hash

        self.__send(response)

    def __send(self,message,secure=False):
        # If the  connection to wattpilot is over a unsecure channel (http) all send messages are wrapped in
        # a "securedMsg" Message which contains the original messageobject and a sha256 HMAC Hashed created
        # using the password
        if secure:
            messageid=message["requestId"]
            payload=json.dumps(message)
            h = hmac.new(bytearray(self._hashedpassword), bytearray(payload.encode()), hashlib.sha256 )
            message={}
            message["type"]="securedMsg"
            message["data"]=payload
            message["requestId"]=str(messageid)+"sm"
            message["hmac"]=h.hexdigest()

        _LOGGER.debug("Message send: %s",json.dumps(message)  )
        self._wsapp.send(json.dumps(message))

    def __on_AuthSuccess(self,message):
        self._connected = True
        _LOGGER.info("Authentication successful")

    def __on_FullStatus(self,message):
        props = message.status.__dict__
        for key in props:
            self.__update_property(key,props[key])

    def __on_AuthError(self,message):
        if message.message=="Wrong password":
            self._wsapp.close()
            _LOGGER.error("Authentication failed: %s" , message.message)

    def __on_DeltaStatus(self,message):
        self._allPropsInitialized=True # Assume all properties have been initialized when first delta status is received
        props = message.status.__dict__
        for key in props:
            self.__update_property(key,props[key])


    def __on_clearInverters(self,message):
        pass

    def __on_updateInverter(self,message):
        pass

    def __on_response(self,message):
        if message.success:
            props = message.status.__dict__
            for key in props:
                self.__update_property(key,props[key])
        else:
            _LOGGER.error("Error Sending Request %s. Message: %s" ,message.requestId,message.message)

    def __on_error(self,wsapp,err):
        self._wsapp.close()
        self._connected=False
        sleep(30)
        self._wsapp.run_forever()

    def __on_close(self,wsapp,code,msg):
        self._connected=False

    def __on_message(self, wsapp, message):
        ## called whenever a message through websocket is received
        _LOGGER.debug("Message received: %s", message)
        msg=json.loads(message, object_hook=lambda d: SimpleNamespace(**d))
        if (msg.type == 'hello'):  # Hello Message -> Received upon connection before auth
            self.__on_hello(msg)
        if (msg.type == 'authRequired'): # Auth Required -> Received after hello 
            self.__on_auth(wsapp,msg)
        if (msg.type == 'response'): # Response Message -> Received after sending a update and contains result of update
            self.__on_response(msg)
        if (msg.type == 'authSuccess'): #Auth Success -> Received after sending  correct authentication message
            self.__on_AuthSuccess(msg)
        if (msg.type == 'authError'): # AUth errot -> Received after sending incorrect authentication message (e.g. wrong password)
            self.__on_AuthError(msg)
        if (msg.type == 'fullStatus'): #Full Status -> Received after successfull connection. Contains all Properties of Wattpilot
            self.__on_FullStatus(msg)
        if (msg.type == 'deltaStatus'): # Delta Status -> Whenever a property changes a Delta Status is send
            self.__on_DeltaStatus(msg)
        if (msg.type == 'clearInverters'): # Unknown
            self.__on_clearInverters(msg)
        if (msg.type == 'updateInverter'): # Contains information of connected Photovoltaik inverter / powermeter
            self.__on_updateInverter(msg)
        if self._message_callback != None:
            self._message_callback(self,wsapp,msg,message)


    def __init__(self, ip ,password,serial=None,cloud=False):
        RECONNECTINTERVALL = 30

        self.__requestid=0
        self._name = None
        self._hostname = None
        self._friendlyName = None
        self._manufacturer = None
        self._devicetype = None
        self._protocol = None
        self._secured = None
        self._serial = None
        self._password = None

        self.password = password

        if(cloud):
            self._url= "wss://app.wattpilot.io/app/" + serial + "?version=1.2.9"
        else:
            self._url = "ws://"+ip+"/ws"
        self.serial = None
        self._connected = False
        self._allProps={}
        self._allPropsInitialized=False
        self._voltage1=None
        self._voltage2=None
        self._voltage3=None
        self._voltageN=None
        self._amps1=None
        self._amps2=None
        self._amps3=None
        self._power1=None
        self._power2=None
        self._power3=None
        self._powerN=None
        self._power=None
        self._version = None
        self._amp = None
        self._AccessState = None
        self._firmware = None
        self._WifiSSID = None
        self._AllowCharging = None
        self._mode=None
        self._carConnected=None
        self._cae=None
        self._cak=None
        self._message_callback=None
        self._property_callback=None

        self._wst=threading.Thread()

        websocket.setdefaulttimeout(10)
        self._wsapp = websocket.WebSocketApp(self.url, on_message=self.__on_message, on_error=self.__on_error, on_close=self.__on_close)
        _LOGGER.info ("Wattpilot %s initilized",self.serial)




import cmd
import json
import logging
import math
import os
import paho.mqtt.client as mqtt
import re
import sys
import wattpilot
import yaml
import pkgutil

from importlib.metadata import version
from time import sleep
from threading import Event
from types import SimpleNamespace

_LOGGER = logging.getLogger(__name__)


#### Utility Functions ####

def utils_add_to_dict_unique(d, k, v):
    if k in d:
        _LOGGER.warning(
            f"About to add duplicate key {k} to dictionary - skipping!")
    else:
        d[k] = v
    return d


def utils_wait_timeout(fn, timeout):
    """Generic timeout waiter"""
    t = 0
    within_timeout = True
    while not fn() and t < timeout:
        sleep(1)
        t += 1
    if t >= timeout:
        within_timeout = False
    return within_timeout


class JSONNamespaceEncoder(json.JSONEncoder):
    # See https://gist.github.com/jdthorpe/313cafc6bdaedfbc7d8c32fcef799fbf
    def default(self, obj):
        if isinstance(obj, SimpleNamespace):
            return obj.__dict__
        return super(JSONNamespaceEncoder, self).default(obj)


def utils_value2json(value):
    return json.dumps(value, cls=JSONNamespaceEncoder)


#### Wattpilot Functions ####

def wp_read_apidef():
    global WATTPILOT_SPLIT_PROPERTIES

    api_definition = pkgutil.get_data(__name__, "ressources/wattpilot.yaml")
    wpdef = {
        "config": {},
        "messages": {},
        "properties:": {},
        "splitProperties": [],
    }
    try:
        wpdef["config"] = yaml.safe_load(api_definition)
        wpdef["messages"] = dict(zip(
            [x["key"] for x in wpdef["config"]["messages"]],
            [x for x in wpdef["config"]["messages"]],
        ))
        wpdef["properties"] = {}
        for p in wpdef["config"]["properties"]:
            wpdef["properties"] = utils_add_to_dict_unique(
                wpdef["properties"], p["key"], p)
            if "childProps" in p and WATTPILOT_SPLIT_PROPERTIES:
                for cp in p["childProps"]:
                    cp = {
                        # Defaults for split properties:
                        "description": f"This is a child property of '{p['key']}'. See its description for more information.",
                        "category": p["category"] if "category" in p else "",
                        "jsonType": p["itemType"] if "itemType" in p else "",
                    } | cp | {
                        # Overrides for split properties:
                        "parentProperty": p["key"],
                        "rw": "R",  # NOTE: Split properties currently can only be read
                    }
                    _LOGGER.debug(f"Adding child property {cp['key']}: {cp}")
                    wpdef["properties"] = utils_add_to_dict_unique(
                        wpdef["properties"], cp["key"], cp)
                    wpdef["splitProperties"].append(cp["key"])
        _LOGGER.debug(
            f"Resulting properties config:\n{utils_value2json(wpdef['properties'])}")
    except yaml.YAMLError as exc:
        _LOGGER.fatal(exc)
    return wpdef


def wp_initialize(host, password):
    global wp
    # Connect to Wattpilot:
    wp = wattpilot.Wattpilot(host, password)
    wp.connect()
    # Wait for connection and initialization:
    utils_wait_timeout(lambda: wp.connected, WATTPILOT_CONNECT_TIMEOUT) or exit(
        "ERROR: Timeout while connecting to Wattpilot!")
    utils_wait_timeout(lambda: wp.allPropsInitialized, WATTPILOT_INIT_TIMEOUT) or exit(
        "ERROR: Timeout while waiting for property initialization!")
    return wp


def wp_get_child_prop_value(cp):
    global wpdef
    cpd = wpdef["properties"][cp]
    if "parentProperty" not in cpd:
        _LOGGER.warning(
            f"Child property '{cpd['key']}' is not linked to a parent property: {cpd}")
        return None
    ppd = wpdef["properties"][cpd["parentProperty"]]
    parent_value = wp.allProps[ppd["key"]]
    value = None
    if ppd["jsonType"] == "array":
        value = parent_value[int(cpd["valueRef"])] if int(
            cpd["valueRef"]) < len(parent_value) else None
        _LOGGER.debug(f"  -> got array value {value}")
    elif ppd["jsonType"] == "object":
        if parent_value == None:
            value = None
            _LOGGER.debug(f"  -> parent value is None, so child as well")
        elif isinstance(parent_value, SimpleNamespace) and cpd["valueRef"] in parent_value.__dict__:
            value = parent_value.__dict__[cpd["valueRef"]]
            _LOGGER.debug(f"  -> got object value {value}")
        elif cpd["valueRef"] in parent_value:
            value = parent_value[cpd["valueRef"]]
            _LOGGER.debug(f"  -> got object value {value}")
        else:
            _LOGGER.warning(
                f"Unable to map child property {cpd['key']}: type={type(parent_value)}, value={utils_value2json(parent_value)}")
    else:
        _LOGGER.warning(f"Property {ppd['key']} cannot be split!")
    return value


def wp_get_all_props(available_only=True):
    global WATTPILOT_SPLIT_PROPERTIES
    global wp
    global wpdef
    if available_only:
        props = {k: v for k, v in wp.allProps.items()}
        if WATTPILOT_SPLIT_PROPERTIES:
            for cp_key in wpdef["splitProperties"]:
                props[cp_key] = wp_get_child_prop_value(cp_key)
    else:
        props = {k: (wp.allProps[k] if k in wp.allProps else None)
                 for k in wpdef["properties"].keys()}
    return props


#### Shell Functions ####

class WattpilotShell(cmd.Cmd):
    intro = f"Welcome to the Wattpilot Shell {version('wattpilot')}.   Type help or ? to list commands.\n"
    prompt = 'wattpilot> '
    file = None
    watching_messages = []
    watching_properties = []

    def postloop(self) -> None:
        print()
        return super().postloop()

    def emptyline(self) -> bool:
        return False

    def _complete_list(self, clist, text):
        return [x for x in clist if x.startswith(text)]

    def _complete_message(self, text, sender=None):
        global wpdef
        return [md["key"] for md in wpdef["messages"].values() if (not sender or md["sender"] == sender) and md["key"].startswith(text)]

    def _complete_propname(self, text, rw=False, available_only=True):
        global wpdef
        return [k for k in wp_get_all_props(available_only).keys() if (not rw or ("rw" in wpdef["properties"][k] and wpdef["properties"][k]["rw"] == "R/W")) and k.startswith(text)]

    def _complete_values(self, text, line):
        global wpdef
        token = line.split(' ')
        if len(token) == 2:
            return self._complete_propname(text, rw=False, available_only=True) + ['<propRegex>']
        elif len(token) == 3 and text in wpdef["properties"]:
            return ['<value>', '<valueRegex>']
        return []

    def do_EOF(self, arg: str) -> bool | None:
        """Exit the shell"""
        return True

    def do_connect(self, arg: str) -> bool | None:
        """Connect to Wattpilot (using WATTPILOT_* env variables)
Usage: connect"""
        global WATTPILOT_HOST
        global WATTPILOT_PASSWORD
        global wp
        wp = wp_initialize(WATTPILOT_HOST, WATTPILOT_PASSWORD)

    def do_exit(self, arg: str) -> bool | None:
        """Exit the shell
Usage: exit"""
        return True

    def do_get(self, arg: str) -> bool | None:
        """Get a property value
Usage: get <propName>"""
        global wp
        global wpdef
        args = arg.split(' ')
        if not self._ensure_connected():
            return
        if len(args) < 1 or arg == '':
            print(f"ERROR: Wrong number of arguments!")
        elif args[0] in wp.allProps:
            pd = wpdef["properties"][args[0]]
            print(mqtt_get_encoded_property(pd, wp.allProps[args[0]]))
        elif args[0] in wpdef["splitProperties"]:
            pd = wpdef["properties"][args[0]]
            print(mqtt_get_encoded_property(
                pd, wp_get_child_prop_value(pd["key"])))
        else:
            print(f"ERROR: Unknown property: {args[0]}")

    def complete_get(self, text, line, begidx, endidx):
        return self._complete_propname(text, rw=False, available_only=True)

    def do_ha(self, arg: str) -> bool | None:
        """Control Home Assistant discovery (+MQTT client)
Usage: ha <enable|disable|discover|properties|start|status|stop|undiscover> [args...]

Home Assistant commands:
  enable <propName>
    Enable a discovered entity representing the property <propName>
    NOTE: Re-enabling of disabled entities may still be broken in HA and require a restart of HA.
  disable <propName>
    Disable a discovered entity representing the property <propName>
  discover <propName>
    Let HA discover an entity representing the property <propName>
  properties
    List properties activated for HA discovery
  start
    Start HA MQTT discovery (using HA_* env variables)
  status
    Status of HA MQTT discovery
  stop
    Stop HA MQTT discovery
  undiscover <propName>
    Let HA remove a discovered entity representing the property <propName>
    NOTE: Removing of disabled entities may still be broken in HA and require a restart of HA.
"""
        global HA_ENABLED
        global HA_PROPERTIES
        global mqtt_client
        args = arg.split(' ')
        if not self._ensure_connected():
            return
        if len(args) < 1 or arg == '':
            print(f"ERROR: Wrong number of arguments!")
            return
        if args[0] == "properties":
            print(
                f"List of properties activated for discovery: {HA_PROPERTIES}")
        elif args[0] == "start":
            HA_ENABLED = 'true'
            mqtt_client = ha_setup(wp)
        elif args[0] == "stop":
            ha_stop(mqtt_client)
            HA_ENABLED = 'false'
        elif args[0] == "status":
            print(
                f"HA discovery is {'enabled' if HA_ENABLED == 'true' else 'disabled'}.")
        elif len(args) > 1 and args[0] in ['enable', 'disable', 'discover', 'undiscover']:
            self._ha_prop_cmds(args[0], args[1])
        else:
            print(f"ERROR: Unsupported argument: {args[0]}")

    def _ha_prop_cmds(self, cmd, prop_name):
        global HA_PROPERTIES
        global MQTT_PROPERTIES
        global mqtt_client
        global wp
        global wpdef
        if prop_name not in wpdef["properties"]:
            print(f"ERROR: Unknown property '{prop_name}!")
        elif cmd == "enable":
            if prop_name not in MQTT_PROPERTIES:
                MQTT_PROPERTIES.append(prop_name)
            ha_discover_property(
                wp, mqtt_client, wpdef["properties"][prop_name], disable_discovery=False, force_enablement=True)
        elif cmd == "disable":
            if prop_name in MQTT_PROPERTIES:
                MQTT_PROPERTIES.remove(prop_name)
            ha_discover_property(
                wp, mqtt_client, wpdef["properties"][prop_name], disable_discovery=False, force_enablement=False)
        elif cmd == "discover":
            if prop_name not in HA_PROPERTIES:
                HA_PROPERTIES.append(prop_name)
            if prop_name not in MQTT_PROPERTIES:
                MQTT_PROPERTIES.append(prop_name)
            ha_discover_property(
                wp, mqtt_client, wpdef["properties"][prop_name], disable_discovery=False, force_enablement=True)
        elif cmd == "undiscover":
            if prop_name in HA_PROPERTIES:
                HA_PROPERTIES.remove(prop_name)
            if prop_name in MQTT_PROPERTIES:
                MQTT_PROPERTIES.remove(prop_name)
            ha_discover_property(
                wp, mqtt_client, wpdef["properties"][prop_name], disable_discovery=True, force_enablement=False)

    def complete_ha(self, text, line, begidx, endidx):
        token = line.split(' ')
        if len(token) == 2:
            return self._complete_list(['enable', 'disable', 'discover', 'properties', 'start', 'status', 'stop', 'undiscover'], text)
        elif len(token) == 3 and token[1] == 'discover':
            return self._complete_list([p for p in self._complete_propname(text, available_only=True) if p not in HA_PROPERTIES], text)
        elif len(token) == 3 and token[1] in ['enable', 'disable', 'undiscover']:
            return self._complete_list(HA_PROPERTIES, text)
        return []

    def do_info(self, arg: str) -> bool | None:
        """Print device infos
Usage: info"""
        global wp
        if not self._ensure_connected():
            return
        print(wp)

    def do_mqtt(self, arg: str) -> bool | None:
        """Control the MQTT bridge
Usage: mqtt <publish|start|status|stop|unpublish> [args...]

MQTT commands:
  properties
    List properties activated for MQTT publishing
  publish <messages|properties>
    Enable publishing of messages or properties
  publish <message> <msgType>
    Enable publishing of a certain message type
  publish <property> <propName>
    Enable publishing of a certain property
  start
    Start the MQTT bridge (using MQTT_* env variables)
  status
    Status of the MQTT bridge
  stop
    Stop the MQTT bridge
  unpublish <messages|properties>
    Disable publishing of messages or properties
  unpublish <message> <msgType>
    Disable publishing of a certain message type
  unpublish <property> <propName>
    Disable publishing of a certain property
"""
        global MQTT_ENABLED
        global mqtt_client
        global wp
        args = arg.split(' ')
        if not self._ensure_connected():
            return
        if len(args) < 1 or arg == '':
            print(f"ERROR: Wrong number of arguments!")
            return
        if args[0] == "properties":
            print(
                f"List of properties activated for MQTT publishing: {MQTT_PROPERTIES}")
        elif args[0] == "start":
            MQTT_ENABLED = 'true'
            mqtt_client = mqtt_setup(wp)
        elif args[0] == "stop":
            mqtt_stop(mqtt_client)
            MQTT_ENABLED = 'false'
        elif args[0] == "status":
            print(
                f"MQTT client is {'enabled' if MQTT_ENABLED == 'true' else 'disabled'}.")
        elif len(args) > 1 and args[0] in ['publish', 'unpublish']:
            self._mqtt_prop_cmds(args[0], args[1])
        else:
            print(f"ERROR: Unsupported argument: {args[0]}")

    def _mqtt_prop_cmds(self, cmd, prop_name):
        global MQTT_PROPERTIES
        global mqtt_client
        global wp
        global wpdef
        if prop_name not in wpdef["properties"]:
            print(f"ERROR: Undefined property '{prop_name}'!")
        elif cmd == "publish" and prop_name not in MQTT_PROPERTIES:
            MQTT_PROPERTIES.append(prop_name)
        elif cmd == "unpublish" and prop_name in MQTT_PROPERTIES:
            MQTT_PROPERTIES.remove(prop_name)

    def complete_mqtt(self, text, line, begidx, endidx):
        token = line.split(' ')
        if len(token) == 2:
            return self._complete_list(['properties', 'publish', 'start', 'status', 'stop', 'unpublish'], text)
        elif len(token) == 3 and token[1] == 'publish':
            return self._complete_list([p for p in self._complete_propname(text, available_only=True) if p not in MQTT_PROPERTIES], text)
        elif len(token) == 3 and token[1] == 'unpublish':
            return self._complete_list(MQTT_PROPERTIES, text)
        return []

    def do_properties(self, arg: str) -> bool | None:
        """List property definitions and values
Usage: properties [propRegex]"""
        global wpdef
        if not self._ensure_connected():
            return
        props = self._get_props_matching_regex(arg, available_only=False)
        if not props:
            print(f"No matching properties found!")
            return
        print(f"Properties:")
        for prop_name, value in sorted(props.items()):
            self._print_prop_info(wpdef["properties"][prop_name], value)
        print()

    def complete_properties(self, text, line, begidx, endidx):
        return self._complete_propname(text, rw=False, available_only=False) + ['<propRegex>']

    def do_rawvalues(self, arg: str) -> bool | None:
        """List raw values of properties (without value mapping)
Usage: rawvalues [propRegex] [valueRegex]"""
        global wp
        if not self._ensure_connected():
            return
        print(f"List raw values of properties (without value mapping):")
        props = self._get_props_matching_regex(arg)
        for pd, value in sorted(props.items()):
            print(f"- {pd}: {utils_value2json(value)}")
        print()

    def complete_rawvalues(self, text, line, begidx, endidx):
        return self._complete_values(text, line)

    def do_server(self, arg: str) -> bool | None:
        """Start in server mode (infinite wait loop)
Usage: server"""
        if not self._ensure_connected():
            return
        _LOGGER.info("Server started.")
        try:
            Event().wait()
        except KeyboardInterrupt:
            _LOGGER.info("Server shutting down.")
        return True

    def do_set(self, arg: str) -> bool | None:
        """Set a property value
Usage: set <propName> <value>"""
        global wp
        global wpdef
        args = arg.split(' ')
        if not self._ensure_connected():
            return
        if len(args) < 2 or arg == '':
            print(f"ERROR: Wrong number of arguments!")
        elif args[0] not in wp.allProps:
            print(f"ERROR: Unknown property: {args[0]}")
        else:
            if args[1].lower() in ["false", "true"]:
                v = json.loads(args[1].lower())
            elif str(args[1]).isnumeric():
                v = int(args[1])
            elif str(args[1]).isdecimal():
                v = float(args[1])
            else:
                v = str(args[1])
            wp.send_update(args[0], mqtt_get_decoded_property(
                wpdef["properties"][args[0]], v))

    def complete_set(self, text, line, begidx, endidx):
        global wpdef
        token = line.split(' ')
        if len(token) == 2:
            return self._complete_propname(text, rw=True, available_only=True)
        elif len(token) == 3 and token[1] in wpdef["properties"]:
            pd = wpdef["properties"][token[1]]
            if "jsonType" in pd and pd["jsonType"] == 'boolean':
                return [v for v in ['false', 'true'] if v.startswith(text)]
            elif "valueMap" in pd:
                return [v for v in pd["valueMap"].values() if v.startswith(text)]
            elif "jsonType" in pd:
                return [f"<{pd['jsonType']}>"]
        return []

    def do_unwatch(self, arg: str) -> bool | None:
        """Unwatch a message or property
Usage: unwatch <message|property> <msgType|propName>"""
        global wp
        args = arg.split(' ')
        if not self._ensure_connected():
            return
        if len(args) < 2 or arg == '':
            print(f"ERROR: Wrong number of arguments!")
        elif args[0] == 'message' and args[1] not in self.watching_messages:
            print(f"ERROR: Message of type '{args[1]}' is not watched")
        elif args[0] == 'message':
            self.watching_messages.remove(args[1])
            if len(self.watching_messages) == 0:
                wp.unregister_message_callback()
        elif args[0] == 'property' and args[1] not in self.watching_properties:
            print(f"ERROR: Property with name '{args[1]}' is not watched")
        elif args[0] == 'property':
            self.watching_properties.remove(args[1])
            if len(self.watching_properties) == 0:
                wp.unregister_property_callback()
        else:
            print(f"ERROR: Unknown watch type: {args[0]}")

    def complete_unwatch(self, text, line, begidx, endidx):
        token = line.split(' ')
        if len(token) == 2:
            return self._complete_list(['message', 'property'], text)
        elif len(token) == 3 and token[1] == 'message':
            return self._complete_list(self.watching_messages, text)
        elif len(token) == 3 and token[1] == 'property':
            return self._complete_list(self.watching_properties, text)
        return []

    def do_values(self, arg: str) -> bool | None:
        """List values of properties (with value mapping enabled)
Usage: values [propRegex] [valueRegex]"""
        global wp
        global wpdef
        if not self._ensure_connected():
            return
        print(f"List values of properties (with value mapping):")
        props = self._get_props_matching_regex(arg)
        for pd, value in sorted(props.items()):
            print(
                f"- {pd}: {mqtt_get_encoded_property(wpdef['properties'][pd],value)}")
        print()

    def complete_values(self, text, line, begidx, endidx):
        return self._complete_values(text, line)

    def do_watch(self, arg: str) -> bool | None:
        """Watch message or a property
Usage: watch <message|property> <msgType|propName>"""
        global wp
        global wpdef
        args = arg.split(' ')
        if not self._ensure_connected():
            return
        if len(args) < 2 or arg == '':
            print(f"ERROR: Wrong number of arguments!")
        elif args[0] == 'message' and args[1] not in wpdef['messages']:
            print(f"ERROR: Unknown message type: {args[1]}")
        elif args[0] == 'message':
            msg_type = args[1]
            if len(self.watching_messages) == 0:
                wp.register_message_callback(self._watched_message_received)
            if msg_type not in self.watching_messages:
                self.watching_messages.append(msg_type)
        elif args[0] == 'property' and args[1] not in wp.allProps:
            print(f"ERROR: Unknown property: {args[1]}")
        elif args[0] == 'property':
            prop_name = args[1]
            if len(self.watching_properties) == 0:
                wp.register_property_callback(self._watched_property_changed)
            if prop_name not in self.watching_properties:
                self.watching_properties.append(prop_name)
        else:
            print(f"ERROR: Unknown watch type: {args[0]}")

    def complete_watch(self, text, line, begidx, endidx):
        global wpdef
        token = line.split(' ')
        if len(token) == 2:
            return self._complete_list(['message', 'property'], text)
        elif len(token) == 3 and token[1] == 'message':
            return self._complete_message(text, 'server')
        elif len(token) == 3 and token[1] == 'property':
            return self._complete_propname(text, rw=False, available_only=True) + ['<propRegex>']
        return []

    def _print_prop_info(self, pd, value):
        global wp
        _LOGGER.debug(f"Property definition: {pd}")
        title = ""
        desc = ""
        alias = ""
        rw = ""
        if 'alias' in pd:
            alias = f", alias:{pd['alias']}"
        if 'rw' in pd:
            rw = f", rw:{pd['rw']}"
        if 'title' in pd:
            title = pd['title']
        if 'description' in pd:
            desc = pd['description']
        print(f"- {pd['key']} ({pd['jsonType']}{alias}{rw}): {title}")
        if desc:
            print(f"  Description: {desc}")
        if pd['key'] in wp.allProps.keys():
            print(
                f"  Value: {mqtt_get_encoded_property(pd,value)}{' (raw:' + utils_value2json(value) + ')' if 'valueMap' in pd else ''}")
        else:
            print(
                f"  NOTE: This property is currently not provided by the connected device!")

    def _watched_property_changed(self, name, value):
        global wpdef
        if name in self.watching_properties:
            pd = wpdef["properties"][name]
            _LOGGER.info(
                f"Property {name} changed to {mqtt_get_encoded_property(pd,value)}")

    def _watched_message_received(self, wp, wsapp, msg, msg_json):
        if msg.type in self.watching_messages:
            _LOGGER.info(f"Message of type {msg.type} received: {msg}")

    def _ensure_connected(self):
        global wp
        if not wp:
            print('Not connected to wattpilot!')
            return False
        return True

    def _get_props_matching_regex(self, arg, available_only=True):
        global wp
        global wpdef
        args = arg.split(' ')
        prop_regex = '.*'
        if len(args) > 0 and args[0] != '':
            prop_regex = args[0]
        props = {k: v for k, v in wp_get_all_props(available_only).items() if re.match(
            r'^'+prop_regex+'$', k, flags=re.IGNORECASE)}
        value_regex = '.*'
        if len(args) > 1:
            value_regex = args[1]
        props = {k: v for k, v in props.items() if re.match(r'^'+value_regex+'$',
                                                            str(mqtt_get_encoded_property(wpdef["properties"][k], v)), flags=re.IGNORECASE)}
        return props


#### MQTT Functions ####

def mqtt_get_mapped_value(pd, value):
    mapped_value = value
    if value == None:
        mapped_value = None
    elif "valueMap" in pd:
        if str(value) in list(pd["valueMap"].keys()):
            mapped_value = pd["valueMap"][str(value)]
        else:
            _LOGGER.warning(
                f"Unable to map value '{value}' of property '{pd['key']} - using unmapped value!")
    return mapped_value


def mqtt_get_mapped_property(pd, value):
    if value and "jsonType" in pd and pd["jsonType"] == "array":
        mapped_value = []
        for v in value:
            mapped_value.append(mqtt_get_mapped_value(pd, v))
    else:
        mapped_value = mqtt_get_mapped_value(pd, value)
    return mapped_value


def mqtt_get_remapped_value(pd, mapped_value):
    remapped_value = mapped_value
    if "valueMap" in pd:
        if mapped_value in pd["valueMap"].values():
            remapped_value = json.loads(str(list(pd["valueMap"].keys())[
                                        list(pd["valueMap"].values()).index(mapped_value)]))
        else:
            _LOGGER.warning(
                f"Unable to remap value '{mapped_value}' of property '{pd['key']} - using mapped value!")
    return remapped_value


def mqtt_get_remapped_property(pd, mapped_value):
    if "jsonType" in pd and pd["jsonType"] == "array":
        remapped_value = []
        for v in mapped_value:
            remapped_value.append(mqtt_get_remapped_value(pd, v))
    else:
        remapped_value = mqtt_get_remapped_value(pd, mapped_value)
    return remapped_value


def mqtt_get_encoded_property(pd, value):
    mapped_value = mqtt_get_mapped_property(pd, value)
    if value == None or "jsonType" in pd and (
            pd["jsonType"] == "array"
            or pd["jsonType"] == "object"
            or pd["jsonType"] == "boolean"):
        return json.dumps(mapped_value, cls=JSONNamespaceEncoder)
    else:
        return mapped_value


def mqtt_get_decoded_property(pd, value):
    if "jsonType" in pd and (pd["jsonType"] == "array" or pd["jsonType"] == "object"):
        decoded_value = json.loads(value)
    else:
        decoded_value = value
    return mqtt_get_remapped_property(pd, decoded_value)


def mqtt_publish_property(wp, mqtt_client, pd, value, force_publish=False):
    prop_name = pd["key"]
    if not (force_publish or MQTT_PROPERTIES == [''] or prop_name in MQTT_PROPERTIES):
        _LOGGER.debug(f"Skipping publishing of property '{prop_name}' ...")
        return
    property_topic = mqtt_subst_topic(MQTT_TOPIC_PROPERTY_STATE, {
        "baseTopic": MQTT_TOPIC_BASE,
        "serialNumber": wp.serial,
        "propName": prop_name,
    })
    encoded_value = mqtt_get_encoded_property(pd, value)
    _LOGGER.debug(
        f"Publishing property '{prop_name}' with value '{encoded_value}' to MQTT ...")
    mqtt_client.publish(property_topic, encoded_value, retain=True)
    if WATTPILOT_SPLIT_PROPERTIES and "childProps" in pd:
        _LOGGER.debug(
            f"Splitting child props of property {prop_name} as {pd['jsonType']} for value {value} ...")
        for cpd in pd["childProps"]:
            _LOGGER.debug(f"Extracting child property {cpd['key']},  ...")
            split_value = wp_get_child_prop_value(cpd['key'])
            _LOGGER.debug(
                f"Publishing sub-property {cpd['key']} with value {split_value} to MQTT ...")
            mqtt_publish_property(wp, mqtt_client, cpd, split_value, True)


def mqtt_publish_message(wp, wsapp, msg, msg_json):
    global mqtt_client
    global MQTT_PUBLISH_MESSAGES
    global MQTT_TOPIC_BASE
    global MQTT_PUBLISH_PROPERTIES
    global MQTT_TOPIC_MESSAGES
    global wpdef
    if mqtt_client == None:
        _LOGGER.debug(f"Skipping MQTT message publishing.")
        return
    msg_dict = json.loads(msg_json)
    if MQTT_PUBLISH_MESSAGES == "true" and (MQTT_MESSAGES == [] or MQTT_MESSAGES == [''] or msg.type in MQTT_MESSAGES):
        message_topic = mqtt_subst_topic(MQTT_TOPIC_MESSAGES, {
            "baseTopic": MQTT_TOPIC_BASE,
            "serialNumber": wp.serial,
            "messageType": msg.type,
        })
        mqtt_client.publish(message_topic, msg_json)
    if MQTT_PUBLISH_PROPERTIES == "true" and msg.type in ["fullStatus", "deltaStatus"]:
        for prop_name, value in msg_dict["status"].items():
            pd = wpdef["properties"][prop_name]
            mqtt_publish_property(wp, mqtt_client, pd, value)

# Substitute topic patterns


def mqtt_subst_topic(s, values, expand=True):
    if expand:
        s = re.sub(r'^~', MQTT_TOPIC_PROPERTY_BASE, s)
    all_values = {
        "baseTopic": MQTT_TOPIC_BASE,
    } | values
    return s.format(**all_values)


def mqtt_setup_client(host, port, client_id, available_topic, command_topic):
    # Connect to MQTT server:
    mqtt_client = mqtt.Client(client_id)
    mqtt_client.on_message = mqtt_set_value
    _LOGGER.info(f"Connecting to MQTT host {host} on port {port} ...")
    mqtt_client.will_set(
        available_topic, payload="offline", qos=0, retain=True)
    mqtt_client.connect(host, port)
    mqtt_client.loop_start()
    mqtt_client.publish(available_topic, payload="online", qos=0, retain=True)
    _LOGGER.info(f"Subscribing to command topics {command_topic}")
    mqtt_client.subscribe(command_topic)
    return mqtt_client


def mqtt_setup(wp):
    global MQTT_CLIENT_ID
    global MQTT_HOST
    global MQTT_PORT
    global MQTT_PROPERTIES
    global MQTT_TOPIC_AVAILABLE
    global MQTT_TOPIC_PROPERTY_SET
    # Connect to MQTT server:
    mqtt_client = mqtt_setup_client(MQTT_HOST, MQTT_PORT, MQTT_CLIENT_ID,
                                    mqtt_subst_topic(MQTT_TOPIC_AVAILABLE, {}),
                                    mqtt_subst_topic(MQTT_TOPIC_PROPERTY_SET, {
                                                     "propName": "+"}),
                                    )
    MQTT_PROPERTIES = mqtt_get_watched_properties(wp)
    _LOGGER.info(
        f"Registering message callback to publish updates to the following properties to MQTT: {MQTT_PROPERTIES}")
    wp.register_message_callback(mqtt_publish_message)
    return mqtt_client


def mqtt_stop(mqtt_client):
    if mqtt_client.is_connected():
        _LOGGER.info(f"Disconnecting from MQTT server ...")
        mqtt_client.disconnect()

# Subscribe to topic for setting property values:


def mqtt_set_value(client, userdata, message):
    global wpdef
    topic_regex = mqtt_subst_topic(
        MQTT_TOPIC_PROPERTY_SET, {"propName": "([^/]+)"})
    name = re.sub(topic_regex, r'\1', message.topic)
    if not name or name == "" or not wpdef["properties"][name]:
        _LOGGER.warning(f"Unknown property '{name}'!")
    pd = wpdef["properties"][name]
    if pd['rw'] == "R":
        _LOGGER.warning(f"Property '{name}' is not writable!")
    value = mqtt_get_decoded_property(
        pd, str(message.payload.decode("utf-8")))
    _LOGGER.info(
        f"MQTT Message received: topic={message.topic}, name={name}, value={value}")
    wp.send_update(name, value)


def mqtt_get_watched_properties(wp):
    global MQTT_PROPERTIES
    if MQTT_PROPERTIES == [] or MQTT_PROPERTIES == ['']:
        return list(wp.allProps.keys())
    else:
        return MQTT_PROPERTIES


#### Home Assistant Functions ####

# Generate device information for HA discovery
def ha_get_device_info(wp):
    ha_device = {
        "connections": [
        ],
        "identifiers": [
            f"wattpilot_{wp.serial}",
        ],
        "manufacturer": wp.manufacturer,
        "model": wp.devicetype,
        "name": wp.name,
        "suggested_area": "Garage",
        "sw_version": wp.version,
    }
    if "maca" in wp.allProps:
        ha_device["connections"] += [["mac", wp.allProps["maca"]]]
    if "macs" in wp.allProps:
        ha_device["connections"] += [["mac", wp.allProps["macs"]]]
    return ha_device


def ha_get_component_for_prop(prop_info):
    component = "sensor"
    if "rw" in prop_info and prop_info["rw"] == "R/W":
        if "valueMap" in prop_info:
            component = "select"
        elif "jsonType" in prop_info and prop_info["jsonType"] == "boolean":
            component = "switch"
        elif "jsonType" in prop_info and prop_info["jsonType"] == "float":
            component = "number"
        elif "jsonType" in prop_info and prop_info["jsonType"] == "integer":
            component = "number"
    elif "rw" in prop_info and prop_info["rw"] == "R":
        if "jsonType" in prop_info and prop_info["jsonType"] == "boolean":
            component = "binary_sensor"
    return component


def ha_get_default_config_for_prop(prop_info):
    config = {}
    if "rw" in prop_info and prop_info["rw"] == "R/W":
        if "jsonType" in prop_info and \
                (prop_info["jsonType"] == "float" or prop_info["jsonType"] == "integer"):
            config["mode"] = "box"
        if "category" in prop_info and prop_info["category"] == "Config":
            config["entity_category"] = "config"
    if "homeAssistant" not in prop_info:
        config["enabled_by_default"] = False
    return config


def ha_get_template_filter_from_json_type(json_type):
    template = "{{ value | string }}"
    if json_type == "float":
        template = "{{ value | float }}"
    elif json_type == "integer":
        template = "{{ value | int }}"
    elif json_type == "boolean":
        template = "{{ value == 'true' }}"
    return template

# Publish HA discovery config for a single property


def ha_discover_property(wp, mqtt_client, pd, disable_discovery=False, force_enablement=None):
    global HA_TOPIC_CONFIG
    global WATTPILOT_SPLIT_PROPERTIES
    global MQTT_TOPIC_PROPERTY_BASE
    global MQTT_TOPIC_PROPERTY_SET
    global MQTT_TOPIC_PROPERTY_STATE
    name = pd["key"]
    ha_info = {}
    if "homeAssistant" in pd:
        ha_info = pd["homeAssistant"] or {}
    component = ha_get_component_for_prop(pd)
    if "component" in ha_info:  # Override component from config
        component = ha_info["component"]
    _LOGGER.debug(
        f"Homeassistant config: haInfo={ha_info}, component={component}")
    title = pd.get("title", pd.get("alias", name))
    _LOGGER.debug(
        f"Publishing HA discovery config for property '{name}' ...")
    ha_config = ha_info.get("config", {})
    unique_id = f"wattpilot_{wp.serial}_{name}"
    object_id = f"wattpilot_{name}"
    topic_subst_map = {
        "component": component,
        "propName": name,
        "serialNumber": wp.serial,
        "uniqueId": unique_id,
    }
    ha_device = ha_get_device_info(wp)
    base_topic = mqtt_subst_topic(
        MQTT_TOPIC_PROPERTY_BASE, topic_subst_map, False)
    ha_discovery_config = ha_get_default_config_for_prop(pd) | {
        "~": base_topic,
        "name": title,
        "object_id": object_id,
        "unique_id": unique_id,
        "state_topic": mqtt_subst_topic(MQTT_TOPIC_PROPERTY_STATE, topic_subst_map, False),
        "availability_topic": mqtt_subst_topic(MQTT_TOPIC_AVAILABLE, {}),
        "payload_available": "online",
        "payload_not_available": "offline",
        "device": ha_device,
    }
    if "valueMap" in pd:
        ha_discovery_config["options"] = list(pd["valueMap"].values())
    if pd.get("rw", "") == "R/W":
        ha_discovery_config["command_topic"] = mqtt_subst_topic(
            MQTT_TOPIC_PROPERTY_SET, topic_subst_map, False)
    ha_discovery_config = dict(
        list(ha_discovery_config.items())
        + list(ha_config.items())
    )
    if force_enablement != None:
        ha_discovery_config["enabled_by_default"] = force_enablement
    topic_cfg = mqtt_subst_topic(HA_TOPIC_CONFIG, topic_subst_map)
    if disable_discovery:
        payload = ''
    else:
        payload = utils_value2json(ha_discovery_config)
    _LOGGER.debug(
        f"Publishing property '{name}' to {topic_cfg}: {payload}")
    mqtt_client.publish(topic_cfg, payload, retain=True)
    # Publish additional read-only sensor for special rw properties:
    if pd.get("rw", "") == "R/W" and component != "sensor":
        if payload != "":
            del ha_discovery_config["command_topic"]
            payload = utils_value2json(ha_discovery_config)
        mqtt_client.publish(mqtt_subst_topic(HA_TOPIC_CONFIG, topic_subst_map | {
                            "component": "sensor"}), payload, retain=True)
    if WATTPILOT_SPLIT_PROPERTIES and "childProps" in pd:
        for p in pd["childProps"]:
            ha_discover_property(wp, mqtt_client, p,
                                 disable_discovery, force_enablement)


def ha_is_default_prop(pd):
    global HA_DISABLED_ENTITIES
    v = "homeAssistant" in pd
    if HA_DISABLED_ENTITIES != 'true':
        ha = pd.get("homeAssistant", {}) if pd.get("homeAssistant", {}) else {}
        v = v and ha.get("config", {}).get("enabled_by_default", True)
    return v


def ha_get_discovery_properties():
    global HA_PROPERTIES
    global wpdef
    _LOGGER.debug(
        f"get_ha_discovery_properties(): HA_PROPERTIES='{HA_PROPERTIES}', propdef size='{len(wpdef['properties'])}'")
    ha_properties = HA_PROPERTIES
    if ha_properties == [''] or ha_properties == []:
        ha_properties = [p["key"]
                         for p in wpdef["properties"].values() if ha_is_default_prop(p)]
    _LOGGER.debug(
        f"get_ha_discovery_properties(): ha_properties='{ha_properties}'")
    return ha_properties


def ha_discover_properties(mqtt_client, ha_properties, disable_discovery=True):
    global wpdef
    _LOGGER.info(
        f"{'Disabling' if disable_discovery else 'Enabling'} HA discovery for the following properties: {ha_properties}")
    for name in ha_properties:
        ha_discover_property(
            wp, mqtt_client, wpdef["properties"][name], disable_discovery)


def ha_publish_initial_properties(wp, mqtt_client):
    global HA_PROPERTIES
    global wpdef
    _LOGGER.info(
        f"Publishing all initial property values to MQTT to populate the entity values ...")
    for prop_name in HA_PROPERTIES:
        if prop_name in wp.allProps:
            value = wp.allProps[prop_name]
            pd = wpdef["properties"][prop_name]
            mqtt_publish_property(wp, mqtt_client, pd, value)


def ha_setup(wp):
    global HA_PROPERTIES
    global HA_WAIT_INIT_S
    global HA_WAIT_PROPS_MS
    global MQTT_PROPERTIES
    global wpdef
    # Configure list of relevant properties:
    HA_PROPERTIES = ha_get_discovery_properties()
    if MQTT_PROPERTIES == [] or MQTT_PROPERTIES == ['']:
        MQTT_PROPERTIES = HA_PROPERTIES
    # Setup MQTT client:
    mqtt_client = mqtt_setup(wp)
    # Publish HA discovery config:
    ha_discover_properties(mqtt_client, HA_PROPERTIES, False)
    # Wait a bit for HA to catch up:
    wait_time = math.ceil(
        HA_WAIT_INIT_S + len(HA_PROPERTIES)*HA_WAIT_PROPS_MS*0.001)
    if wait_time > 0:
        _LOGGER.info(
            f"Waiting {wait_time}s to allow Home Assistant to discovery entities and subscribe MQTT topics before publishing initial values ...")
        # Sleep to let HA discover the entities before publishing values
        sleep(wait_time)
    # Publish initial property values to MQTT:
    ha_publish_initial_properties(wp, mqtt_client)
    return mqtt_client


def ha_stop(mqtt_client):
    global HA_PROPERTIES
    ha_discover_properties(mqtt_client, HA_PROPERTIES, True)
    mqtt_stop(mqtt_client)


#### Main Program ####

def main_setup_env():
    global HA_DISABLED_ENTITIES
    global HA_ENABLED
    global HA_PROPERTIES
    global HA_TOPIC_CONFIG
    global HA_WAIT_INIT_S
    global HA_WAIT_PROPS_MS
    global MQTT_AVAILABLE_PAYLOAD
    global MQTT_CLIENT_ID
    global MQTT_ENABLED
    global MQTT_HOST
    global MQTT_MESSAGES
    global MQTT_NOT_AVAILABLE_PAYLOAD
    global MQTT_PORT
    global MQTT_PROPERTIES
    global MQTT_PUBLISH_MESSAGES
    global MQTT_PUBLISH_PROPERTIES
    global MQTT_TOPIC_AVAILABLE
    global MQTT_TOPIC_BASE
    global MQTT_TOPIC_MESSAGES
    global MQTT_TOPIC_PROPERTY_BASE
    global MQTT_TOPIC_PROPERTY_SET
    global MQTT_TOPIC_PROPERTY_STATE
    global WATTPILOT_AUTOCONNECT
    global WATTPILOT_CONNECT_TIMEOUT
    global WATTPILOT_DEBUG_LEVEL
    global WATTPILOT_HOST
    global WATTPILOT_INIT_TIMEOUT
    global WATTPILOT_PASSWORD
    global WATTPILOT_SPLIT_PROPERTIES
    HA_DISABLED_ENTITIES = os.environ.get('HA_DISABLED_ENTITIES', 'false')
    HA_ENABLED = os.environ.get('HA_ENABLED', 'false')
    HA_PROPERTIES = os.environ.get('HA_PROPERTIES', '').split(sep=' ')
    HA_TOPIC_CONFIG = os.environ.get(
        'HA_TOPIC_CONFIG', 'homeassistant/{component}/{uniqueId}/config')
    HA_WAIT_INIT_S = int(os.environ.get('HA_WAIT_INIT_S', '0'))
    HA_WAIT_PROPS_MS = int(os.environ.get('HA_WAIT_PROPS_MS', '0'))
    MQTT_AVAILABLE_PAYLOAD = os.environ.get('MQTT_AVAILABLE_PAYLOAD', 'online')
    MQTT_CLIENT_ID = os.environ.get('MQTT_CLIENT_ID', 'wattpilot2mqtt')
    MQTT_ENABLED = os.environ.get('MQTT_ENABLED', 'false')
    MQTT_HOST = os.environ.get('MQTT_HOST', '')
    MQTT_MESSAGES = os.environ.get('MQTT_MESSAGES', '').split(sep=' ')
    MQTT_NOT_AVAILABLE_PAYLOAD = os.environ.get(
        'MQTT_NOT_AVAILABLE_PAYLOAD', 'offline')
    MQTT_PORT = int(os.environ.get('MQTT_PORT', '1883'))
    MQTT_PROPERTIES = os.environ.get('MQTT_PROPERTIES', '').split(sep=' ')
    MQTT_PUBLISH_MESSAGES = os.environ.get('MQTT_PUBLISH_MESSAGES', 'false')
    MQTT_PUBLISH_PROPERTIES = os.environ.get('MQTT_PUBLISH_PROPERTIES', 'true')
    MQTT_TOPIC_AVAILABLE = os.environ.get(
        'MQTT_TOPIC_AVAILABLE', '{baseTopic}/available')
    MQTT_TOPIC_BASE = os.environ.get('MQTT_TOPIC_BASE', 'wattpilot')
    MQTT_TOPIC_MESSAGES = os.environ.get(
        'MQTT_TOPIC_MESSAGES', '{baseTopic}/messages/{messageType}')
    MQTT_TOPIC_PROPERTY_BASE = os.environ.get(
        'MQTT_TOPIC_PROPERTY_BASE', '{baseTopic}/properties/{propName}')
    MQTT_TOPIC_PROPERTY_SET = os.environ.get(
        'MQTT_TOPIC_PROPERTY_SET', '~/set')
    MQTT_TOPIC_PROPERTY_STATE = os.environ.get(
        'MQTT_TOPIC_PROPERTY_STATE', '~/state')
    WATTPILOT_AUTOCONNECT = os.environ.get('WATTPILOT_AUTOCONNECT', 'true')
    WATTPILOT_CONNECT_TIMEOUT = int(
        os.environ.get('WATTPILOT_CONNECT_TIMEOUT', '30'))
    WATTPILOT_DEBUG_LEVEL = os.environ.get('WATTPILOT_DEBUG_LEVEL', 'INFO')
    WATTPILOT_HOST = os.environ.get('WATTPILOT_HOST', '')
    WATTPILOT_INIT_TIMEOUT = int(
        os.environ.get('WATTPILOT_INIT_TIMEOUT', '30'))
    WATTPILOT_PASSWORD = os.environ.get('WATTPILOT_PASSWORD', '')
    WATTPILOT_SPLIT_PROPERTIES = bool(
        os.environ.get('WATTPILOT_SPLIT_PROPERTIES', 'true'))

    # Ensure wattpilot host an password are set:
    assert WATTPILOT_HOST != '', "WATTPILOT_HOST not set!"
    assert WATTPILOT_PASSWORD != '', "WATTPILOT_PASSWORD not set!"
    assert MQTT_ENABLED == 'false' or MQTT_HOST != '', 'MQTT_HOST not set!'


def main():
    global MQTT_ENABLED
    global WATTPILOT_AUTOCONNECT
    global WATTPILOT_DEBUG_LEVEL
    global mqtt_client
    global wp
    global wpdef

    # Setup environment variables:
    main_setup_env()

    # Set debug level:
    logging.basicConfig(level=WATTPILOT_DEBUG_LEVEL)

    # Initialize globals:
    mqtt_client = None
    wp = None
    wpdef = wp_read_apidef()

    # Initialize shell:
    wpsh = WattpilotShell()
    if WATTPILOT_AUTOCONNECT == 'true':
        _LOGGER.info("Automatically connecting to Wattpilot ...")
        wpsh.onecmd('connect')
        # Enable MQTT and/or HA integration:
        if MQTT_ENABLED == "true" and HA_ENABLED == "false":
            wpsh.onecmd('mqtt start')
        elif MQTT_ENABLED == "true" and HA_ENABLED == "true":
            wpsh.onecmd('ha start')
        wpsh.onecmd('info')
    if len(sys.argv) < 2:
        wpsh.cmdloop()
    else:
        wpsh.onecmd(sys.argv[1])


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)




from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='wattpilot',
    version='0.2',
    description='Python library to connect to a Fronius Wattpilot Wallbox',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/joscha82/wattpilot',
    author='Joscha Arenz',
    author_email='joscha@arenz.co',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3.10",
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='wattpilot',  
    package_dir={'': 'src'}, 
    packages=find_packages(where='src'),
    entry_points = {
        'console_scripts': ['wattpilotshell=wattpilot.wattpilotshell:main'],
    },
    package_data = { '' : ['wattpilot.yaml'] },
    python_requires='>=3.10, <4',
    install_requires=['websocket-client','PyYAML','paho-mqtt'],
    platforms="any",
    license="MIT License",
    project_urls={
        'Bug Reports': 'https://github.com/joscha82/wattpilot/issues',
        'Source': 'https://github.com/joscha82/wattpilot'
    },
    include_package_data=True
)
